/*
License: LGPL
$Id: code-n.c,v 1.11 2011/12/02 06:25:11 taka Exp $
cf. @s/2011/12/01-my-note-code-n.pdf
 */
#include <stdio.h>
#include <stdlib.h> 
#include <math.h>
/* RANK is given by -DRANK=...
   M is given by -DM...  RANK=2^M
 */
int bitcount(int n) {int c; c=0; while (n) { ++c; n &= (n-1); } return(c);}
#define join(k,jj) ( (1 << k) | jj)
#define delete(k,jj) ( (~(1 << k)) & jj)
#define member(k,ii) ( (1 << k) & ii)
#define NaN 9.999e100

/*
  p(x) = exp(-bsum*x)*x^(M*(c-a))
  f is p(x)*pd{J}F
 */
void rf(double x, double *f, int rank_not_used, double *val, int n_not_used)
{ extern double *Beta; /* beta = (1/2)*\simga^(-1) */
  extern double *Ng;   /* freedom */
  extern int Mg;       /* number of variables, Mg=M */
  static int initialized = 0;
  static double b[M];
  static double bsum = 0;
  static double a;
  static double c;
  static int m;
  static int bitSize[RANK];
  static int invbitSize[RANK];
  int i,j,k,p;
  int ii,jj; /* stands for I and J */
  double rijj;
  double yik[M][M];
  double yik2[M][M];
  double y[M];
  double f2[M][RANK];

  if (!initialized) {
    m = Mg;
    if (m != M) error_code("M != m. M is given by -DM...");
    a = ((double)m+1.0)/2.0;

    if(Beta != NULL) for (i=0;i<M;i++) b[i]=Beta[i];
    else error_code("Beta is null.");
    if (Ng != NULL) c=a+*(Ng)/2;  /* set c, a+n/2, n is given by the option n */
    else error_code("Ng is null.");

    bsum = 0; for (i=0; i<m; i++) bsum += b[i];

    for (i=0; i<RANK; i++) bitSize[i] = bitcount(i);
    for (i=0, k=0; i<=M; i++) {
      for (j=0; j<RANK; j++) {
	if (bitSize[j] == i) {invbitSize[k] = j; k++; }
      }
    }
    initialized = 1;
  }
  for (i=0; i<M; i++) {
    y[i] = b[i]*x;    
  }
  for (i=0; i<M; i++) {
    for (k=0; k<M; k++) {
      if (i != k) {
	yik[i][k] = y[k]/(y[i]-y[k]);
	yik2[i][k] = y[i]/((y[i]-y[k])*(y[i]-y[k]));
      }
    }
  }

  /* 
     generation of f2[i][jj] = p(x) pd{i}^2 \pd{J} F 
     Do not forget to devide by y[i]
  */
  for (i=0; i<M; i++) for (j=0; j<RANK; j++) f2[i][j] = NaN;
  for (i=0; i<M; i++) {
    f2[i][0] = -((c-y[i])*f[join(i,0)]-a*f[0]);
    for (k=0; k<M; k++) {
      if (i!=k) {
	f2[i][0] += -0.5*yik[i][k]*(f[join(i,0)]-f[join(k,0)]);
      }
    }
    f2[i][0] = f2[i][0]/y[i];
  }
  for (p=1; p<RANK; p++) {
    jj = invbitSize[p];
    for (i=0; i<M; i++) {
      if (member(i,jj)) continue;
      ii = join(i,jj);
      rijj = -( (c-y[i])*f[ii]-a*f[jj] );
      for (k=0; k<M; k++) {
	if (!member(k,ii)) rijj += -(0.5)*yik[i][k]*(f[ii]-f[join(k,jj)]);
	if (member(k,jj)) {
	  rijj += -(0.5)*yik[i][k]*f[ii];
	  rijj += -(0.5)*yik2[i][k]*(f[join(i,delete(k,jj))]-f[jj]);
	}
      }
      f2[i][jj] = rijj;
      for (k = 0; k<M; k++) {
	if (member(k,jj)) {
	  f2[i][jj] += (0.5)*yik[i][k]*f2[k][delete(k,jj)];
	}
      }
      f2[i][jj] = f2[i][jj]/y[i];
    }
  }
  /** showf2("f2",f2,M,RANK); exit(0);  */

  for (jj=0; jj<RANK; jj++) {
    val[jj] = 0;
    for (i=0; i<M; i++) {
      if (member(i,jj)) {
	val[jj] += b[i]*f2[i][delete(i,jj)];
      }else{
	val[jj] += b[i]*f[join(i,jj)];
      }
    }
  }

   /*   diagonal shift. check M.*/
  for (i=0; i<RANK; i++) {
    val[i] += (-bsum+M*(c-a)/x)*f[i];
  }
}

error_code(char *s) {
  fprintf(stderr,"%s",s);
  exit(10);
}

/* for debug */
showf(char *s,double *v,int m) {
  int i;
  printf("%s=\n",s);
  for (i=0; i<m; i++) {
    printf("%e, ",v[i]);
  }
  printf("\n");
}

showd(char *s,int *v,int m) {
  int i;
  printf("%s=\n",s);
  for (i=0; i<m; i++) {
    printf("%5d, ",v[i]);
  }
  printf("\n");
}

showf2(char *s,double *v,int m,int n) {
  int i,j;
  printf("%s=\n",s);
  for (i=0; i<m; i++) {
    for (j=0; j<n; j++) {
      printf("%e, ",v[i*n+j]);
    }
    printf("\n");
  }
  printf("\n");
  printf("member(0,0)=%d, member(0,1)=%d, member(0,2)=%d,member(0,3)=%d\n",
	 member(0,0),member(0,1),member(0,2),member(0,3));
}

