/*
mh-n.c, linked with jack-n.c 
License: LGPL
$Id: mh-n.c,v 1.8 2011/12/24 10:40:46 taka Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
/* #define M_n 3 */

/* global variables. They are set in setParam() */
int Mg;  /* n */
int Mapprox; /* m, approximation degree */
double *Beta; /* beta[0], ..., beta[m-1] */
double *Ng;   /* freedom n.  c=(m+1)/2+n/2; Note that it is a pointer */
double X0g;   /* initial point */
double *Iv;   /* Initial values of mhg sorted by mhbase() in rd.rr at beta*x0 */
double Ef;   /* exponential factor at beta*x0 */
double Hg;   /* step size of rk defined in rk.c */
int Dp;      /* Data sampling period */
double Xng=0.0;   /* the last point */
extern int Debug;
int Sample = 1;
extern double M_rel_error;
extern int UseTable; /* Use the table for beta or not. The table method is buggy.*/

/* for sample inputs */
double *Iv2; 
double Ef2; 

#define SMAX 4096
#define inci(i) { i++; if (i >= argc) { fprintf(stderr,"Option argument is not given.\n"); return(-1); }}

int gopen_file(void);
int setParam(char *fname);
int showParam(void);
double iv_factor(void);
double gammam(double a,int n);
double mh_t(double a[],double b[],int n,int m);
double mh_t2(int j);
double mypower(double a,int n);
int imypower(int x,int n) {
  int i;
  int v;
  if (n < 0) {myerror("imypower"); exit(-1);}
  v = 1;
  for (i=0; i<n; i++) v *= x;
  return(v);
}
void *mymalloc(int);

main(int argc,char *argv[]) {
  double *y0;
  double x0,xn;
  double ef;
  int i,j,rank;
  double a[1]; double b[1];
  extern double M_x[];
  extern double *Beta;
  extern int M_2n;
  Debug = 0;
  UseTable = 1;
  setParam(NULL);  Mapprox=6;
  for (i=1; i<argc; i++) {
	if (strcmp(argv[i],"--idata")==0) {
	  inci(i);
	  setParam(argv[i]);
	}else if (strcmp(argv[i],"--degree")==0) {
	  inci(i);
	  sscanf(argv[i],"%d",&Mapprox);
	}else if (strcmp(argv[i],"--x0")==0) {
	  inci(i);
	  sscanf(argv[i],"%lg",&X0g);
	}else if (strcmp(argv[i],"--notable")==0) {
	  UseTable = 0;
	}else if (strcmp(argv[i],"--debug")==0) {
	  Debug = 1;
	}else if (strcmp(argv[i],"--help")==0) {
	  usage(); return(0);
	}else {
	  fprintf(stderr,"Unknown option %s\n",argv[i]);
	  usage();
	  return(-1);
	}
  }

  printf("%%%%Use --help option to see the help.\n");
  printf("%%%%Mapprox=%d\n",Mapprox);
  if (M_n != Mg) {
	myerror("Mg must be equal to %d\n",M_n); exit(-1);
  }
  if (Debug) showParam();
  for (i=0; i<M_n; i++) {
	M_x[i] = Beta[i]*X0g;
  } 
  a[0] = ((double)Mg+1.0)/2.0;
  b[0] = ((double)Mg+1.0)/2.0 + ((double) (*Ng))/2.0; /* bug, double check */
  if (Debug) printf("Calling mh_t with ([%lf],[%lf],%d,%d)\n",a[0],b[0],M_n,Mapprox);
  mh_t(a,b,M_n,Mapprox);
  if (imypower(2,M_n) != M_2n) {
	printf("M_n=%d,M_2n=%d\n",M_n,M_2n);
	myerror("2^M_n != M_2n\n"); exit(-1);
  }
  printf("%%%%M_rel_error=%lg\n",M_rel_error);
  for (j=0; j<M_2n; j++) {
	Iv[j] = mh_t2(j);
  }
  Ef = iv_factor();
  showParam();
}

usage() {
  fprintf(stderr,"Usages:\n");
  fprintf(stderr,"mh-m [--idata input_data_file --x0 x0 --degree approxm]\n");
  fprintf(stderr,"\nThe command mh-m [options] generates an input for w-m, Pr({y | y<xmax}), which is the cumulative distribution function of the largest root of the m by m Wishart matrix with n degrees of freedom and the covariantce matrix sigma.\n");
  fprintf(stderr,"The mh-m uses the Koev-Edelman algorithm to evalute the matrix hypergeometric function.\n");
  fprintf(stderr,"The degree of the approximation (Mapprox) is given by the --degree option.\n");
  fprintf(stderr,"Parameters are specified by the input_data_file. Otherwise, default values are used.\n");
  fprintf(stderr,"The format of the input_data_file. The orders of the input numbers must be kept.\n");
  fprintf(stderr," Mg: m(the number of variables), Beta: beta=sigma^(-1)/2 (diagonized), Ng: n,\n");
  fprintf(stderr," Iv: initial values at X0g*Beta (see our paper how to order them), are evaluated in this program. Give zeros. \n");
  fprintf(stderr," Ef: a scalar factor to the initial value. It is calculated by this program. Give the zero.\n");
  fprintf(stderr," Hg: h (step size) which is for w-m, X0g: starting value of x(when --x0 option is used, this value is used), Xng: terminating value of x which is for w-m.\n");
  fprintf(stderr," Dp: output data is stored in every Dp steps when output_data_file is specified, which is for w-m.\n");
  fprintf(stderr," The line started with %% is a comment line.\n");
  fprintf(stderr," With the --notable option, it does not use the Lemma 3.2 of Koev-Edelman (there is a typo: kappa'_r = mu'_r for 1<=r<=mu_k).\n");
  fprintf(stderr," An example format of the input_data_file can be obtained by executing mh-m with no option.\n");
  fprintf(stderr,"\nExamples:\n");
  fprintf(stderr,"[1] ./mh-2 \n");
  fprintf(stderr,"[2] ./mh-2 --x0 0.3 \n");
  fprintf(stderr,"[3] ./mh-3 --x0 0.1 \n");
  fprintf(stderr,"[4] ./mh-3 --x0 0.1 --degree 15 \n");
  fprintf(stderr,"[5] ./mh-3 --idata test.txt --degree 15 \n");
  fprintf(stderr,"[6] ./mh-3 --degree 15 >test2.txt\n");
  fprintf(stderr,"    ./w-3 --idata test2.txt --gnuplotf test-g\n");
  fprintf(stderr,"    gnuplot -persist <test-g-gp.txt\n");
}

setParamDefault() {
  int rank;
  int i;
  Mg = M_n ;
  rank = imypower(2,Mg);
  Beta = (double *)mymalloc(sizeof(double)*Mg);
  for (i=0; i<Mg; i++) Beta[i] = 1.0+i;
  Ng = (double *)mymalloc(sizeof(double)); *Ng = 3.0;
  Iv = (double *)mymalloc(sizeof(double)*rank);
  Iv2 = (double *)mymalloc(sizeof(double)*rank);
  for (i=0; i<rank; i++) Iv[i] = 0;
  Ef = 0;
  Ef2 = 0.01034957388338225707;
  if (M_n == 2) {
    Iv2[0] = 1.58693;
    Iv2[1] = 0.811369;
    Iv2[2] = 0.846874;
    Iv2[3] = 0.413438;
  }
  X0g = (Beta[0]/Beta[Mg-1])*0.5;
  Hg = 0.001;
  Dp = 1;
  Xng = 10.0;
}

next(FILE *fp,char *s,char *msg) {
  s[0] = '%';
  while (s[0] == '%') {
	if (!fgets(s,SMAX,fp)) {
	  fprintf(stderr,"Data format error at %s\n",msg);
	  exit(-1);
	}
	if (s[0] != '%') return(0);
  }
}
setParam(char *fname) {
  int rank;
  char s[SMAX];
  FILE *fp;
  int i;
  if (fname == NULL) return(setParamDefault());

  Sample = 0;
  if ((fp=fopen(fname,"r")) == NULL) {
	fprintf(stderr,"File %s is not found.\n",fname);
	exit(-1);
  }
  next(fp,s,"Mg(m)");
  sscanf(s,"%d",&Mg);
  rank = imypower(2,Mg);

  Beta = (double *)mymalloc(sizeof(double)*Mg);
  for (i=0; i<Mg; i++) {
    next(fp,s,"Beta");
	sscanf(s,"%lf",&(Beta[i]));
  }

  Ng = (double *)mymalloc(sizeof(double));
  next(fp,s,"Ng(freedom parameter n)");
  sscanf(s,"%lf",Ng);
  
  next(fp,s,"X0g(initial point)");
  sscanf(s,"%lf",&X0g);
  
  Iv = (double *)mymalloc(sizeof(double)*rank);
  for (i=0; i<rank; i++) {
	next(fp,s,"Iv(initial values)");
	sscanf(s,"%lg",&(Iv[i]));
  }

  next(fp,s,"Ef(exponential factor)");
  sscanf(s,"%lg",&Ef);

  next(fp,s,"Hg (step size of rk)");
  sscanf(s,"%lf",&Hg);

  next(fp,s,"Dp (data sampling period)");
  sscanf(s,"%d",&Dp);

  next(fp,s,"Xng (the last point, cf. --xmax)");
  sscanf(s,"%lf",&Xng);
  fclose(fp);
}

showParam() {
  int rank,i;
  rank = imypower(2,Mg);
  printf("%%Mg=\n%d\n",Mg);
  for (i=0; i<Mg; i++) {
	printf("%%Beta[%d]=\n%lf\n",i,Beta[i]);
  }
  printf("%%Ng=\n%lf\n",*Ng);
  printf("%%X0g=\n%lf\n",X0g);
  for (i=0; i<rank; i++) {
	printf("%%Iv[%d]=\n%lg\n",i,Iv[i]);
	if (Sample && (M_n == 2) && (X0g == 0.3))
	  printf("%%Iv[%d]-Iv2[%d]=%lg\n",i,i,Iv[i]-Iv2[i]);
  }
  printf("%%Ef=\n%lg\n",Ef);
  printf("%%Hg=\n%lf\n",Hg);
  printf("%%Dp=\n%d\n",Dp);
  printf("%%Xng=\n%lf\n",Xng);
}

double gammam(double a,int n) {
  double v,v2;
  int i;
  v=mypower(sqrt(M_PI),(n*(n-1))/2);
  v2=0;
  for (i=1; i<=n; i++) {
    v2 += lgamma(a-((double)(i-1))/2.0); /* not for big n */
  }
  if (Debug) printf("gammam(%lg,%d)=%lg\n",a,n,v*exp(v2));
  return(v*exp(v2));
}

double iv_factor(void) {
  double v1;
  double t;
  double b;
  double detSigma;
  double c;
  int i,n;
  n = (int) (*Ng);
  v1= mypower(sqrt(X0g),n*M_n);
  t = 0.0;
  for (i=0; i<M_n; i++)  t += -X0g*Beta[i];
  v1 = v1*exp(t);

  b = 1; for (i=0; i<M_n; i++) b *= Beta[i];
  detSigma = 1.0/(b*mypower(2.0,M_n));

  c = gammam(((double)(M_n+1))/2.0, M_n)/
     ( mypower(sqrt(2), M_n*n)*mypower(sqrt(detSigma),n)*gammam(((double)(M_n+n+1))/2.0,M_n) );
  return( c*v1);
}


