/*
License: LGPL
$Id: rk.c,v 1.16 2011/12/25 00:41:30 taka Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#ifndef RANK
#include "tmp-rank.h"
#endif

char *Gfname;
char *Dfname;
extern int Display;
extern int Mg;
extern double *Beta; extern double *Ng; extern double X0g;
extern int RawName;
double Hg = 0.001;
FILE *Gf=NULL;
FILE *Df =NULL;
int P95=0;
int Verbose=0;
static mypower(int x,int n) {
  int a,i;
  a = 1;
  for (i=0; i<n; i++) a = a*x;
  return(a);
}
gopen_file() {
  FILE *fp;
  char fname[1024];
  int i;
  Gf=NULL;
  Df=NULL;
  if (Verbose) {
    fprintf(stderr,"Gfname=%s\n",Gfname);
    fprintf(stderr,"Dfname=%s\n",Dfname);
  }
  if (RANK != mypower(2,Mg)) {
	fprintf(stderr,"rk.c RANK=%d is not equal to 2^Mg=%d. input_data_file is broken.\n",RANK,mypower(2,Mg));
	usage();
	exit(-1);
  }
  if (Dfname != NULL) {
    if (!RawName) {
      sprintf(&(Dfname[strlen(Dfname)]),"-m%d-n%lf-x0%lf-b",Mg,*Ng,X0g);
      for (i=0; i<Mg; i++) {
	sprintf(&(Dfname[strlen(Dfname)]),"%lf,",Beta[i]);
      }
      sprintf(&(Dfname[strlen(Dfname)]),".txt");
    }
    Df = fopen(Dfname,"w");
    if (Df == NULL) {
      fprintf(stderr,"Error to open the file %s\n",Dfname);
      return(-1);
    }
  }
  if (Gfname != NULL) {
    Gf = fopen(Gfname,"w");
    if (Gf == NULL) {
      fprintf(stderr,"Error to open the file %s\n",Gfname);
      return(-1);
    }
  }else return(1);
  sprintf(fname,"%s-gp.txt",Gfname);
  fp = fopen(fname,"w");
  fprintf(fp,"set xrange [0:20]\n");
  fprintf(fp,"set yrange [0:1.2]\n");
  fprintf(fp,"plot \"%s\" title \"by hg\" with lines\n",Gfname);
  fclose(fp);
  return(0);
}
/*
        Runge-Kutta
        y_{n+1} = y_n + 1/6 k_1 + 1/3 k_2 + 1/3 k_3 + 1/6 k_4
        k_1 = h f(x_n, y_n)
        k_2 = h f(x_n + 1/2 h, y_n + 1/2 k_1)
        k_3 = h f(x_n + 1/2 h, y_n + 1/2 k_2)
        k_4 = h f(x_n + h, y_n + k_3)
*/

/* $B%Y%/%H%kCM4X?t(B rk is defined in tmp-code-?.c,
 f(x, y) (x : scalar, y : vector) */ 

void show_v(double x,double *v, int n)
{
  int i;
  static int counter=0;
  extern int Dp;

  if ((counter % Dp) != 0) { counter++; return;} else counter=1;
  fprintf(Df,"%lf\n",x);
  for (i = 0; i < n; i++) fprintf(Df," %le\n", v[i]);
}

double rkmain(double x0,double y0[],double xn)
{
  int i;
  double h;
  double x; 
  double y[RANK];
  double k1[RANK], k2[RANK], k3[RANK], k4[RANK];
  double temp[RANK];
  double ty[RANK];
  h = Hg;
  for (i = 0; i < RANK; i++)
    y[i] = y0[i];
  for (x = x0; (h>0?(x<xn):(x>xn)); x += h) {
	if (Df) show_v(x,y, RANK);
	if (Gf) fprintf(Gf,"%lf %le\n",x,y[0]);
        /* Output 95% point */
	if (P95) {
	  if ((P95==1) && (y[0] >= 0.95)) {
	    printf("x=%le, y[0]=%lf\n",x,y[0]);
	    P95=2;
	  }else if ((P95==2) && (y[0] >=0.9500000001)) {
	    printf("x=%le, y[0]=%lf\n",x,y[0]);
	    P95=0;
	  }
	}
    rf(x, y, RANK, temp, RANK);
    for (i = 0; i < RANK; i++)
      k1[i] = h * temp[i];

    for (i = 0; i < RANK; i++)
      ty[i] = y[i] + 0.5 * k1[i];
    rf(x + 0.5 * h, ty, RANK, temp, RANK);
    for (i = 0; i < RANK; i++)
      k2[i] = h * temp[i];

    for (i = 0; i < RANK; i++)
      ty[i] = y[i] + 0.5 * k2[i];
    rf(x + 0.5 * h, ty, RANK, temp, RANK);
    for (i = 0; i < RANK; i++)
      k3[i] = h * temp[i];

    for (i = 0; i < RANK; i++)
      ty[i] = y[i] + k3[i];
    rf(x + h, ty, RANK, temp, RANK);
    for (i = 0; i < RANK; i++)
      k4[i] = h * temp[i];
                
    for (i = 0; i < RANK; i++)
      y[i] = y[i] + 1.0/6.0 * k1[i] + 1.0/3.0 * k2[i] + 1.0/3.0 * k3[i] + 1.0/6.0 * k4[i];
  }
  printf("x=%lf, y[0]=%lg\n",x,y[0]);
  if (Df) fclose(Df);
  if (Gf) fclose(Gf);
  return x;
}
